/*
Sound & Light Machine
Firmware
for use with ATtiny2313
Make Magazine issue #10
Mitch Altman
19-Mar-07

modified for mega8 by cm@tahina.priv.at 20071208
$Id: slm.c,v 1.11 2007/12/23 20:38:53 cm Exp cm $
*/

#define F_CPU 8000000UL // 8 MHz

// audio channels:
#define AU_BASE 1
#define AU_PLAY 0

#define BASE_BEAT 200.0 // base frequency

#include <avr/io.h>             // this contains all the IO port definitions
#include <avr/interrupt.h>      // definitions for interrupts
#include <avr/sleep.h>          // definitions for power-down modes
#include <avr/pgmspace.h>       // definitions or keeping constants in program memory

#include "dds.h"

// tuning change
volatile int32_t tune_step;
volatile uint32_t tune_accum;
volatile uint32_t tune_duration;

#define TUNE_BEEP DDS_TUNE(1000.0)


#define DEBOUNCE_COUNT 1000
volatile uint8_t mode;

/*
The hardware for this project is very simple:
     ATtiny2313 has 20 pins:
       pin 1   connects to serial port programming circuitry
       pin 10  ground
       pin 12  PB0 - Left eye LED1
       pin 13  PB1 - Right eye LED1
       pin 14  OC0A - Left ear speaker (base-frequency)
       pin 15  OC1A - Right ear speaker (Offset Frequencies for binaural beats)
       pin 17  connects to serial port programming circuitry
       pin 18  connects to serial port programming circuitry
       pin 19  connects to serial port programming circuitry
       pin 20  +3v
    All other pins are unused

    This firmware requires that the clock frequency of the ATtiny 
      is the default that it is shipped with:  8.0MHz

    The mega8 version has the following pins used:
    PD2(INT0) - switch to GND
    PD6 - LED1
    PD7 - LED2
    PB1 (OC1A) - audio1
    PB2 (OC1B) - audio2
*/

// be careful
#define BWSTEP(s,e,d) \
	(int32_t)((0.0 + DDS_TUNE(e) - DDS_TUNE(s)) * 4096.0 / ((d)*DDS_F_S))

#define DURATION(d) (uint32_t)((d) * DDS_F_S)

// new data structure 
struct bwentry {
  uint32_t start;		// tune at begin of interval
  int32_t step;			// tune step
  uint32_t duration;		// duration of interval, times DDS_F_S
};

// selected sequence
struct bwentry const *sel_bwtab;

#include "bwtab.h"


// This function starts the Offset Frequency audio in the Right ear to
// create a binaural beat (between Left and Right ears) for a
// Brainwave Element (the base-frequency of 400Hz is already assumed
// to be playing in the Left ear before calling this function) and
// blinks the LEDs at the same frequency for the Brainwave Element and
// keeps it going for the Duration specified for the Brainwave Element
static uint8_t do_brainwave_element(int index) {
  uint32_t start = pgm_read_dword(&sel_bwtab[index].start);
  int32_t step = pgm_read_dword(&sel_bwtab[index].step);
  uint32_t duration = pgm_read_dword(&sel_bwtab[index].duration);

  if(start == 0) return 0;

  dds_tune[AU_PLAY] = start;
  tune_accum = start<<12;
  // set tune step last to avoid crash when an interrupt comes at the wrong
  // time
  tune_step = step;
  tune_duration = duration;
  while(tune_duration && mode);	// wait for freq change to complete

  return 1;
}

inline uint8_t keypressed(void) {
  int i = DEBOUNCE_COUNT;
  
  if(PIND & _BV(PD2)) return 0; // key not pressed

  while(i) {
    while(i && !(PIND & _BV(PD2))) i--;
    if(i)
      i = DEBOUNCE_COUNT;
  } // we have a stable key event now
  while(!(PIND & _BV(PD2)));	// wait for release
  return 1;
}

void beep(uint32_t duration) {
  tune_step = 0;
  tune_duration = duration;
  dds_tune[AU_PLAY] = TUNE_BEEP;
  tune_accum = TUNE_BEEP<<12;
  while(tune_duration && mode);
}

int main(void) {

  while(1) {
    GICR &= ~(_BV(INT0)); // disable INT0
    mode = 1; // powered on
    // TIMSK = 0x00;  // no Timer interrupts enabled, default
    DDRB = _BV(DDB1) | // PB1 = OC1A, audio PWM
      _BV(DDB2); // PB2 = OC1B, audio PWM
    PORTB = ~(_BV(PB1) | _BV(PB2));  // all PORTB output pins off,
    // pullups on for inputs

    DDRD = _BV(DDD6) | _BV(DDD7); // PD6, PD7 (LEDs) as outputs
    PORTD = ~(_BV(PD6) | _BV(PD7));  // all PORTD output pins off,
    // pullups for inputs

    // setup timer1 for PWM output with a frequency of DDS_F_S
    //TCNT1 = 0; //default
    //OCR1A = 0;
    //OCR1B = 0;

    TCCR1A = 0xa0;
    // COM1A = 10 clear on match while up-counting, set when down-counting
    // COM1B = 10
    // FOC1A = 0 unused
    // FOC1B = 0
    // WGM11:0 = (10)00 phase + freq correct PWM, TOP = ICR1
    // test wgm11:0 (10)10 phase correct... -> 0xa2

    TCCR1B = 0x11;
    // ICNC1 = 0 unused
    // ICES1 = 0 unused
    // reserved 0
    // WGM13:2 10(00) see above
    // CS12:0 001 prescaler = 1

    TIMSK |= 0x04; // TOIE1: enable overflow interrupt
    ICR1 = 255;

    // play base freq on left channel
    dds_tune[AU_BASE] = DDS_TUNE(BASE_BEAT);

    // select first sequence (for now)
    memcpy_P(&sel_bwtab, &bwtab[0], sizeof(PGM_VOID_P));

    // enable interrupts
    sei();

    //beep(DURATION(0.5));

    // loop through entire Brainwave Table of Brainwave Elements
    int j = 0;
    while(do_brainwave_element(j) && mode) j++;

    // Shut down everything and put the CPU to sleep
    TCCR1B &= ~0x07; // CS12:0: stop clock
    PORTB=0;
    DDRB=0; // all inputs
    PORTD=0;
    DDRD=0; // all inputs
    PORTD=_BV(PD2); // pullup for button
    MCUCR &= ~(_BV(ISC01) | _BV(ISC00)); // low level on INT0
    GICR |= _BV(INT0); // enable INT0
    //cli();  // disable interrupts
    MCUCR |= (_BV(SE) |		// sleep enable
	      _BV(SM1));	// power down sleep mode
    sleep_cpu();           // put CPU into Power Down Sleep Mode
  }
  return 0;			/* notreached */
}


/* gets called on every timer1 overflow, and sets up the next sample to be
   output */
ISR(SIG_OVERFLOW1) {
  // shift frequency
  tune_accum += tune_step;
  dds_tune[AU_PLAY] = tune_accum>>12;
  tune_duration--;

  // set up next sample for the audio channels
  OCR1A = dds(AU_BASE);
  OCR1B = dds(AU_PLAY);

  // blink the LEDs in sync with the binaural beat: if the phase accumulators
  // of the audio channels differ by more than 180, turn leds off, else on.
  if((dds_pa[AU_PLAY] - dds_pa[AU_BASE]) & 0x2000000)
    PORTD = (_BV(PD2));
  else
    PORTD = (_BV(PD6)|_BV(PD7)|_BV(PD2));

  // read key, debounce, reset mode on end of keypress
  if(keypressed())
    mode = 0;
}

// pressing the pushbutton brings us here
ISR(INT0_vect) {
  if(keypressed())
    mode = 1;
}

/*

  $Log: slm.c,v $
  Revision 1.11  2007/12/23 20:38:53  cm
  added an on/off key

  Revision 1.10  2007/12/23 15:29:06  cm
  new data structure for sequence...

  Revision 1.9  2007/12/12 22:18:15  cm
  flash LEDs in phase with binaural beats
  shift frequencies instead of switching

  Revision 1.8  2007/12/12 20:40:09  cm
  cleanup

  Revision 1.7  2007/12/11 23:28:03  cm
  cleaned up DDS

  Revision 1.6  2007/12/11 22:38:17  cm
  optimized

  Revision 1.5  2007/12/09 14:36:32  cm
  table-driven generation...

  Revision 1.4  2007/12/09 14:01:07  cm
  changed timings to seconds

  Revision 1.3  2007/12/09 13:56:55  cm
  forgot the master counter

  Revision 1.2  2007/12/09 13:51:04  cm
  changed to interrupt driven


 */
